# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (c) 2014 E-MIPS (http://www.e-mips.com.ar)
#    Copyright (c) year Aconcagua Team (http://www.proyectoaconcagua.com.ar)
#    All Rights Reserved. See AUTHORS for details.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General
#    Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv, fields
from openerp.tools.translate import _


class ServiceComponent(osv.osv):
    _name = "service.component"
    _rec_name = "product_id"

    def _get_component_ids_partner_related(self, cr, uid, ids, context=None):
        result = {}
        for system in self.pool.get('service.system').browse(
                cr, uid, ids, context=context):
            result.update({
                comp.id: True for comp in system.component_ids
            })
        return result.keys()

    def _set_partner_id(self, cr, uid, ids, name, value, arg, context=None):
        res = {}
        if not isinstance(ids, list):
            ids = [ids]
        for comp in self.browse(cr, uid, ids, context=context):
            if bool(comp.system_id) and (
                    (not comp.system_id.partner_id and value) or
                    comp.system_id.partner_id.id != value):
                raise osv.except_osv(
                    _("Error!"),
                    _("You can not change the Partner of a Component that " +
                      "is related to a System."))
            query = """
            UPDATE service_component SET partner_id = %(partner_id)s
            WHERE id = %(component_id)s
            """
            q_vals = {
                'partner_id': value,
                'component_id': comp.id,
            }
            cr.execute(query, q_vals)
            res[comp.id] = True
        return res

    def _compute_partner_id(self, cr, uid, ids, name, arg, context=None):
        res = {}
        for comp in self.browse(cr, uid, ids, context=context):
            res[comp.id] = comp.system_id and comp.system_id.partner_id and \
                comp.system_id.partner_id.id
        return res

    _columns = {
        'product_id': fields.many2one('product.product',
                                      string="Product",
                                      required=True),
        'partner_id': fields.function(
            _compute_partner_id,
            relation='res.partner', type='many2one',
            fnct_inv=_set_partner_id,
            string='Customer', store={
                'service.component': (lambda s, cr, uid, ids, ctx={}: ids,
                                      ['system_id'], 20),
                'service.system': (_get_component_ids_partner_related,
                                   ['partner_id'], 20),
            }),
        'prodlot_id': fields.many2one('stock.production.lot',
                                      'Serial Number'),
        'system_id': fields.many2one('service.system',
                                     'System'),
        'sale_order_line': fields.many2one('sale.order.line',
                                           'Sale Order Line'),
        'move_id': fields.many2one('stock.move',
                                   'Stock Move', ondelete='restrict'),
        'in_workshop': fields.boolean('In Workshop'),
        'id_equipo': fields.integer("Id CM"),
        'module_id': fields.many2one('service.module', 'Module'),
        'install_date': fields.date(string='Install Date'),
        'in_disuse': fields.boolean(string='In Disuse'),
    }
