# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (c) 2014 E-MIPS (http://www.e-mips.com.ar)
#    Copyright (c) year Aconcagua Team (http://www.proyectoaconcagua.com.ar)
#    All Rights Reserved. See AUTHORS for details.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General
#    Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
from datetime import datetime, timedelta
from itertools import groupby

import netsvc
import pytz
from booking_chart.mixin import mixin
from lxml import etree
from numpy import busday_count
from openerp.tools import DEFAULT_SERVER_DATE_FORMAT as DSDF  # , float_compare
from openerp.tools import DEFAULT_SERVER_DATETIME_FORMAT as DSDTF
from openerp import exceptions
from osv import fields, osv
from tools.translate import _


class service_finish_type(osv.osv):
    _name = "service.finish.type"
    _description = "Type of Service Finish"

    _columns = {
        'name': fields.char("Name", required=True),
    }


# Service Order
class service_order(osv.osv):
    _name = "service.order"
    _inherit = ['mail.thread', 'ir.needaction_mixin']
    _description = "Service Order"
    _order = 'name desc'

    _track = {
        'state': {
            'service.mt_service_order_confirmed':
                lambda self, cr, uid, obj, ctx=None:
                obj['state'] in ['confirmed'],
            'service.mt_service_order_in_progress':
                lambda self, cr, uid, obj, ctx=None:
                obj['state'] in ['in_progress'],
            'service.mt_service_order_ready':
                lambda self, cr, uid, obj, ctx=None: obj['state'] in ['ready'],
            'service.mt_service_order_resolved':
                lambda self, cr, uid, obj, ctx=None:
                obj['state'] in ['resolved'],
            'service.mt_service_order_done':
                lambda self, cr, uid, obj, ctx=None: obj['state'] in ['done'],
            'service.mt_service_order_customer_confirmed':
                lambda self, cr, uid, obj, ctx=None:
                obj['state'] in ['customer_confirmed'],
            'service.mt_service_order_cancel':
                lambda self, cr, uid, obj, ctx=None:
                obj['state'] in ['cancel'],
        },
    }

    def fields_view_get(self, cr, uid, view_id=None,
                        view_type=False, context=None,
                        toolbar=False, submenu=False):
        res = super(service_order, self).fields_view_get(
            cr, uid, view_id=view_id, view_type=view_type,
            context=context, toolbar=toolbar, submenu=submenu)
        user_obj = self.pool.get('res.users')
        if view_type == 'form' and user_obj.has_group(
                cr, uid, 'service.group_service_manager'):
            doc = etree.XML(res['arch'])
            for node in doc.xpath("//field[@name='type']"):
                node.set('modifiers', '{"required": true}')
            res['arch'] = etree.tostring(doc)
        return res

    # Obtiene las esperas asociadas
    def _get_wait(self, cr, uid, ids, context=None):
        service_order_obj = self.pool.get('service.order')
        service_order_ids = service_order_obj.search(
            cr, uid, [('wait_ids', 'in', ids)], context=context)
        return service_order_ids

    # Obtiene las acciones asociadas
    def _get_action(self, cr, uid, ids, context=None):
        service_order_obj = self.pool.get('service.order')
        service_order_ids = service_order_obj.search(
            cr, uid, [('action_ids', 'in', ids)], context=context)
        return service_order_ids

    def _get_qty_action_waits(self, cr, uid, ids,
                              field_name, arg, context=None):
        values = {}
        for service_id in self.browse(cr, uid, ids, context=context):
            # Search quantity waits and actions pending
            qty_act_waits = 0
            for action_id in service_id.action_ids:
                if action_id.state == 'draft':
                    qty_act_waits += 1

            for wait_id in service_id.wait_ids:
                if wait_id.state == 'waiting':
                    qty_act_waits += 1

            values[service_id.id] = qty_act_waits

        return values

    def _get_next_assignation(self, cr, uid, ids,
                              field_name, arg, context=None):
        """ Finds next assignation date
        @return: Dictionary of values
        """
        res = {}
        for id in ids:
            res[id] = False
        if not ids:
            return res
        cr.execute("""SELECT order_id, date
                      FROM
                        service_order_assignee
                      WHERE
                        order_id IN %s
                      AND
                        date>=now()
                      ORDER BY
                        date
                      LIMIT 1""", (tuple(ids),))

        for order_id, date_next in cr.fetchall():
            res[order_id] = date_next
        return res

    def _get_hours_attended(self, cr, uid, ids,
                            field_name, arg, context=None):
        orders = self.browse(cr, uid, ids, context=context)
        res = {}
        for order in orders:
            creation = order.creation_date
            attended = order.attended
            if not creation or not attended:
                res[order.id] = 0
                continue

            waits = filter(
                lambda w: (w.customer_wait and w.date_end and
                           w.date_end <= attended),
                order.wait_ids
            )

            creation = datetime.strptime(creation, DSDTF)
            attended = datetime.strptime(attended, DSDTF)
            res[order.id] = self._hours_between_dates(cr, uid, ids, creation,
                                                      attended, waits)

            # Quizas conviene que los negativos se pongan como 0
            # hours = self._hours_between_dates(cr, uid, ids, creation,
            #                                           attended, waits)
            # res[order.id] = hours >= 0 and hours or 0.0

        return res

    def _get_hours_resolved(self, cr, uid, ids,
                            field_name, arg, context=None):
        orders = self.browse(cr, uid, ids, context=context)
        res = {}
        for order in orders:
            creation = order.creation_date
            resolved = order.date_resolved
            if not creation or not resolved:
                res[order.id] = 0
                continue

            waits = filter(
                lambda w: (w.customer_wait and w.date_end and
                           w.date_end <= resolved),
                order.wait_ids
            )

            creation = datetime.strptime(creation, DSDTF)
            resolved = datetime.strptime(resolved, DSDTF)
            res[order.id] = self._hours_between_dates(cr, uid, ids, creation,
                                                      resolved, waits)

            # Quizas conviene poner los negativos como 0
            # hours = self._hours_between_dates(cr, uid, ids, creation,
            #                                           resolved, waits)
            # res[order.id] = hours >= 0 and hours or 0.0

        return res

    _colors_mapping = {
        'draft': 'black',
        'confirmed': 'black',
        'in_progress': 'light-red',
        'ready': 'orange',
        'customer_confirmed': 'red',
        'resolved': 'black',
        'done': 'black',
        'cancel': 'black',
    }

    _columns = {
        'name': fields.char('Number', size=32, required=True, readonly=True,
                            states={'draft': [('readonly', False)]},
                            select=True),
        'component_ids': fields.many2many(
            'service.component', 'component_service_order_rel',
            'service_order_id', 'component_id', 'Components'),
        'system_id': fields.many2one('service.system', 'System'),
        'system_type_id': fields.related('system_id', 'system_type_id',
                                         type="many2one", string="System Type",
                                         relation="service.system.type",
                                         readonly=True),
        'creation_date': fields.datetime('Creation date', readonly=True),
        'expected_date': fields.date('Expected date', readonly=True,
                                     states={
                                         'draft': [('readonly', False)],
                                         'confirmed': [('readonly', False)],
                                         'in_progress': [('readonly', False)]
                                     }),
        'to_check': fields.boolean('To Check'),
        'component_no_apply': fields.boolean(
            'NA Components',
            help="If checked, it is not necesary to fill " +
            "components to resolve this order.",
            readonly=False,
            states={
                'resolved': [('readonly', True)],
                'done': [('readonly', True)],
                'cancel': [('readonly', True)],
             }),
        'partner_id': fields.many2one('res.partner', 'Partner', readonly=True,
                                      domain=[('parent_id', '=', False),
                                              ('is_company', '=', True)],
                                      states={'draft': [('readonly', False)]},
                                      required=True),
        'sector_id': fields.many2one('res.sector', 'Sector', readonly=True,
                                     states={
                                         'draft': [('readonly', False)],
                                         'confirmed': [('readonly', False)],
                                         'in_progress': [('readonly', False)]
                                     }),
        'employee_sector_id': fields.many2one(
            'res.employee', 'Contact', readonly=True,
            states={
                'draft': [('readonly', False)],
                'confirmed': [('readonly', False)],
                'in_progress': [('readonly', False)]
            }),
        'type': fields.many2one('service.order.type', 'Type', readonly=False,
                states={
                    'resolved': [('readonly', True)],
                    'done': [('readonly', True)],
                    'cancel': [('readonly', True)]
                }),
        'resolved_by_phone': fields.boolean('Resolved By Phone',
                                            readonly=True),
        'quoted': fields.boolean('Quoted', readonly=True,
                                 help="If not set, user should create Quotations or Sales Orders before continue"),  # noqa
        'state': fields.selection([('draft', 'Draft'),
                                   ('confirmed', 'Initiated'),
                                   ('in_progress', 'In Progress'),
                                   ('ready', 'Planificated'),
                                   ('customer_confirmed', 'Confirmed'),
                                   ('resolved', 'Resolved'),
                                   ('done', 'Finished'),
                                   ('cancel', 'Cancelled')], 'State',
                                  readonly=True, select=True),
        # Falta el domain para permitir seleccionar solamente ingenieros.
        'attended': fields.datetime('Attended', readonly=True),
        'attended_by': fields.many2one('res.users', 'Attended by',
                                       readonly=False,
                                       states={
                                           'done': [('readonly', True)]
                                       }),
        'assigned_adm_to': fields.many2one('res.users', 'Responsible',
                                           readonly=False,
                                           states={
                                               'done': [('readonly', True)]
                                           }),
        'assigned_to_ids': fields.one2many('service.order.assignee',
                                           'order_id', 'Assigned to',
                                           readonly=True,
                                           states={
                                               'draft': [('readonly', False)],
                                               'confirmed': [('readonly',
                                                              False)],
                                               'in_progress': [('readonly',
                                                                False)]
                                           }),
        'assigned_by_id': fields.many2one('res.users', 'Planificated by',
                                          readonly=True),
        'assign_date': fields.date('Assign date', readonly=True),
        'finished_by_id': fields.many2one('hr.employee', 'Finished by',
                                          states={
                                              'done': [('readonly', True)]
                                          }),
        'material_line_ids': fields.one2many('service.order.material.line',
                                             'order_id', 'Materials',
                                             states={
                                                 'done': [('readonly', True)]
                                             }),
        'material_line_used_ids': fields.one2many(
            'service.order.material.line.used', 'order_id', 'Materials Used',
            states={'done': [('readonly', True)]}),
        'sale_order_ids': fields.many2many(
            'sale.order', 'service_sale_order_rel', 'sale_order_id',
            'service_order_id', 'Sale Orders',
            states={'done': [('readonly', True)]}),
        'initial_survey': fields.text('Initial survey', readonly=False,
                                      states={
                                          'resolved': [('readonly', True)],
                                          'done': [('readonly', True)],
                                          'cancel': [('readonly', True)]
                                      }),
        'final_diagnostic': fields.text('Final diagnostic'),
        'comments': fields.text('Comments'),
        'communication': fields.text('Communication'),
        'date_resolved': fields.datetime('Resolved date',
                                         states={
                                             'done': [('readonly', True)],
                                         }),
        'date_next_assignee': fields.function(_get_next_assignation,
                                              string='Next Asignation',
                                              type='datetime', readonly=True),
        'wait_ids': fields.one2many('service.wait', 'order_id', 'Waits',
                                    states={'done': [('readonly', True)]}),
        'action_ids': fields.one2many('service.action', 'order_id', 'Actions',
                                      states={'done': [('readonly', True)]}),
        # TODO: Explicar que esta relacion depende si quieren o no un historico de que se le pedia en ese momento que se hizo el service  # noqa
        'requirement_ids': fields.many2many(
            'service.requirement', 'service_order_requirements_rel',
            'order_id', 'requirement_id', "Requirements for Service"),
        'risk_customer': fields.boolean("Risky Customer", readonly=True),
        'parent_id': fields.many2one('service.order', 'Parent Order',
                                     select=True),
        'child_ids': fields.one2many('service.order', 'parent_id',
                                     'Child Orders'),
        'finish_id': fields.many2one('service.finish.type', 'Finish Type'),
        'qty_act_waits': fields.function(
            _get_qty_action_waits, string='Quantity actions and waits',
            type='integer',
            store={
                'service.wait': (_get_wait, ['state'], 20),
                'service.order': (lambda s, cr, uid, ids, c={}:
                                  ids, ['wait_ids', 'action_ids'], 20),
                'service.action': (_get_action, ['state'], 20)
            }),
        'risk': fields.related(
            'partner_id', 'kanban_state', type='selection', string='Risk',
            track_visibility='always', readonly=True,
            selection=[
                ('risk', 'Risk'),
                ('alert', 'Alert'),
                ('normal', 'Normal')],
            ),
        'hours_attended': fields.function(
            _get_hours_attended, type="float",
            string="Quantity of hours to be attended",
            store={
                'service.order': (lambda s, cr, uid, ids, c={}: ids,
                                  ['attended'], 20),
                'service.wait': (_get_wait, ['date_start', 'date_end'], 20),
            }),

        'hours_resolved': fields.function(
            _get_hours_resolved, type="float",
            string="Quantity of hours to be resolved",
            store={
                'service.order': (lambda s, cr, uid, ids, c={}: ids,
                                  ['date_resolved'], 20),
                'service.wait': (_get_wait, ['date_start', 'date_end'], 20),
            }),
    }

    _defaults = {
        'state': 'draft',
        'name': lambda obj, cr, uid, context: '/',
        'quoted': False,
        'to_check': False,
        'creation_date': lambda *x: time.strftime(
            DSDTF),
    }

    def _unificate_overlaps_waits(self, cr, uid, waits, context=None):
        """
        Check if a wait is overlap whit the next and if is overlap,
        it unified them.

        _Return a list of dic whit format:
            [{'date_start': date_start_1, 'date_end': date_end_1},
             {'date_start': date_start_2, 'date_end': date_end_2},
             ...]
        """
        res = []
        if not waits:
            return res

        # Waits sort by date_start
        waits = sorted(waits, key=lambda w: w.date_start)
        ref_start = waits[0].date_start
        ref_end = waits[0].date_end
        for wait in waits[1:]:
            # If Overlaps
            if (ref_start < wait.date_end and
               wait.date_start < ref_end):
                ref_start = min(ref_start, wait.date_start)
                ref_end = max(ref_end, wait.date_end)

            # If not Overlap
            else:
                res.append({
                    'date_start': ref_start,
                    'date_end': ref_end,
                })
                ref_start = wait.date_start
                ref_end = wait.date_end

        # When the loop is over, waits no overlap more, then save ref
        res.append({
            'date_start': ref_start,
            'date_end': ref_end,
        })

        return res

    def _hours_between_dates(self, cr, uid, ids, start_date,
                             end_date, waits=[], context=None):
        """
        Count weekdays in hours between two date
        whitout weekend days, holidays and waits if
        it is provides.
        """

        str_start_date = start_date.date().strftime(DSDF)
        str_end_date = end_date.date().strftime(DSDF)

        # Gets Holidays between dates
        not_assignable_obj = self.pool['not.assignable.times']
        holiday_domain = [('is_holiday', '=', True),
                          ('date_start', '>', str_start_date),
                          ('date_end', '<', str_end_date)]
        holidays_ids = not_assignable_obj.search(cr, uid, holiday_domain,
                                                 context=context)
        holidays = not_assignable_obj.browse(cr, uid, holidays_ids,
                                             context=context)
        holidays = map(
            lambda h: (datetime.strptime(h.date_end, DSDTF),
                       datetime.strptime(h.date_start, DSDTF)),
            holidays
        )
        # Convert to tuples (end_date, start_date) whit datetime obj
        holiday_days = sum(
            (date_end - date_start).days + 1
            for date_end, date_start in holidays
        )

        u_waits = self._unificate_overlaps_waits(cr, uid, waits,
                                                 context=context)
        # Convert to tuples (end_date, start_date) whit datetime obj
        u_waits = map(
            lambda w: (datetime.strptime(w['date_end'], DSDTF),
                       datetime.strptime(w['date_start'], DSDTF)),
            u_waits
        )

        wait_hours = sum(
            self._hours_between_dates(cr, uid, ids, date_start, date_end,
                                      context=context)
            for date_end, date_start in u_waits
        )

        week_mask = '0000011'  # Counting only Saturday and Sunday
        week_mask = list(week_mask)
        if start_date.weekday() == 5 or end_date.weekday() == 5:
            week_mask[5] = '0'
        if start_date.weekday() == 6 or end_date.weekday() == 6:
            week_mask[6] = '0'
        week_mask = ''.join(week_mask)
        weekend_days = busday_count(str_start_date, str_end_date,
                                    weekmask=week_mask)
        hours = (
            (end_date - start_date).total_seconds() / 3600 -
            wait_hours - (holiday_days + weekend_days) * 24
        )
        return hours

    def _check_recursion(self, cr, uid, ids, context=None):
        level = 100
        while len(ids):
            q = """
            SELECT DISTINCT parent_id FROM service_order WHERE id IN %s
            """
            cr.execute(q, (tuple(ids), ))
            ids = filter(None, map(lambda x: x[0], cr.fetchall()))
            if not level:
                return False
            level -= 1
        return True

    _constraints = [
        (_check_recursion, 'Error! You cannot create recursive orders.',
         ['parent_id'])
    ]

    def onchange_partner_id(self, cr, uid, ids, partner_id, sector_id):
        v = {}

        v['sector_id'] = False
        v['employee_sector_id'] = False
        v['system_id'] = False

        if not partner_id:
            return {'value': v}

        partner_obj = self.pool.get('res.partner')
        partner = partner_obj.browse(cr, uid, partner_id)

        v['risk_customer'] = partner.kanban_state == 'risk' and True or False

        domain = {
            'system_id': [('partner_id', '=', partner_id)],
            'component_ids': [('partner_id', '=', partner_id)],
            'sector_id': [('partner_id', '=', partner_id)]
        }
        if sector_id:
            domain['employee_sector_id'] = [
                ('sector_id', '=', sector_id),
                ('partner_id', '=', partner_id)
            ]
        return {
            'value': v,
            'domain': domain
        }

    def onchange_sector_id(self, cr, uid, ids, partner_id, sector_id):
        v = {}

        # Reiniciamos contacto y sistema
        v['employee_sector_id'] = False
        v['system_id'] = False

        domain = {}
        domain_system = [('partner_id', '=', partner_id)]

        if sector_id:
            domain_system.append(('sector_id', '=', sector_id))
            domain['employee_sector_id'] = [
                ('sector_id', '=', sector_id),
                ('partner_id', '=', partner_id)
            ]
        else:
            domain['employee_sector_id'] = [('partner_id', '=', partner_id)]

        if partner_id:
            domain['component_ids'] = [('partner_id', '=', partner_id)]

        domain['system_id'] = domain_system

        return {'value': v, 'domain': domain}

    def onchange_system_id(self, cr, uid, ids, partner_id, system_id):
        v = {}
        if not system_id:
            return {'value': v}

        component_ids = self.pool.get('service.system').read(
            cr, uid, system_id, ['component_ids'])['component_ids']
        d = [('id', 'in', component_ids)]
        if partner_id:
            d.append(('partner_id', '=', partner_id))

        domain = {
            'component_ids': d
        }
        return {
            'value': v,
            'domain': domain
        }

    def create(self, cr, uid, vals, context=None):
        creation_date = resolved_date = datetime.utcnow()
        if vals.get('date_resolved'):
            resolved_date = datetime.strptime(vals.get('date_resolved'), DSDTF)
        self.validate_date_resolver(creation_date, resolved_date)

        partner_obj = self.pool.get('res.partner')

        if 'partner_id' in vals:
            partner = partner_obj.browse(cr, uid, vals['partner_id'])
            vals['risk_customer'] = partner.kanban_state == 'risk' and True \
                or False
        order_id = super(service_order, self).create(
            cr, uid, vals, context=context)
        order = self.browse(cr, uid, order_id, context)
        action_obj = self.pool.get('service.action')
        action_ids = [action_id for action_id in order.action_ids]
        action_obj.validate_date_resolver(cr, uid, action_ids, context)
        return order_id

    def validate_date_resolver(self, creation_date, resolved_date):
        if creation_date > resolved_date:
            raise exceptions.Warning(
                _('La Fecha de Resuelto debe ser posterior a la Fecha de Creación.')
            )

    def check_waits(self, cr, uid, ids, context):
        for service in self.browse(cr, uid, ids, context):
            for wait in service.wait_ids:
                if wait.state != 'finished':
                    raise exceptions.Warning(
                        _('Las esperas deben estar solucionadas antes de resolver el parte')
                    )

    def write(self, cr, uid, ids, vals, context=None):
        if not isinstance(ids, list):
            ids = [ids]
        creation_date = resolved_date = datetime.utcnow()
        for service in self.browse(cr, uid, ids, context):
            creation_date = service.creation_date
            if creation_date:
                creation_date = datetime.strptime(creation_date, DSDTF)
            date_resolved = vals.get('date_resolved')
            if date_resolved:
                resolved_date = datetime.strptime(date_resolved, DSDTF)
            self.validate_date_resolver(creation_date, resolved_date)

        # Actualiza el campo order_state de los assignees
        # en base al state del parte
        # Esto sirve para los colores en el booking chart
        field_list = ['partner_id', 'state', 'system_id',
                      'requirement_ids', 'type']
        need_write = False
        for field in field_list:
            if field in vals:
                need_write = True
        if need_write and ids:
            assignees = []
            for order in self.browse(cr, uid, ids):
                assignees += map(lambda x: x.id, order.assigned_to_ids)
            reads = self.read(cr, uid, ids[0], field_list)

            reqs = ''
            if reads['requirement_ids']:
                requirement_model = self.pool['service.requirement']
                requirements = requirement_model.browse(
                    cr, uid, reads['requirement_ids'], context=context)
                first = True
                for requirement in requirements:
                    if first:
                        reqs += (requirement.name)
                        first = False
                    else:
                        reqs += (' - ' + requirement.name)

            soa_vals = {'requirements': reqs}
            if reads['type']:
                soa_vals['type'] = reads['type'][1]
            if reads['system_id']:
                soa_vals['system_name'] = reads['system_id'][1]

            if 'state' in vals:
                soa_vals['order_state'] = self._colors_mapping[vals['state']]
                if vals.get('state') == 'resolved':
                    self.check_waits(cr, uid, ids, context)
            self.pool.get('service.order.assignee').write(
                cr, uid, assignees, soa_vals)

        if 'assigned_to_ids' in vals:
            vals['assign_date'] = datetime.strftime(
                datetime.today(), DSDTF)
            vals['assigned_by_id'] = uid

        if 'attended_by' in vals:
            services = self.search(cr, uid, [('id', 'in', ids), ('state', '=', 'resolved')])
            if not services:
                vals['attended'] = datetime.strftime(
                    datetime.today(), DSDTF)
        res = super(service_order, self).write(cr, uid, ids, vals, context)
        orders = self.browse(cr, uid, ids, context)
        action_obj = self.pool.get('service.action')
        action_ids = [action_id.id for order in orders for action_id in order.action_ids]
        action_obj.validate_date_resolver(cr, uid, action_ids, context)
        return res

    def _create_stock_move(self, cr, uid, order, material,
                           location_id, context):
        move_obj = self.pool.get('stock.move')
        location_obj = self.pool.get('stock.location')

        res = location_obj.search(cr, uid, [('usage', '=', 'customer')],
                                  limit=1, context=context)

        if not res:
            raise osv.except_osv(
                _("Error in materials"),
                _("You have to configure a Customer Location"))

        # Destino de cliente
        location_dest_id = res[0]

        data = {
            'name': 'Service Order %s' % order.name,
            'origin': order.name,
            'product_id': material.product_id.id,
            'product_qty': material.quantity,
            'product_uom': material.product_uom.id,
            'location_id': location_id,
            'location_dest_id': location_dest_id,
            'state': 'confirmed',
        }

        move_id = move_obj.create(cr, uid, data, context=context)
        return move_id



    def mark_action_done(self, cr, uid, ids, action_id, context=None):
        user_obj = self.pool.get('res.users')
        action_obj = self.pool.get('service.action')

        user = user_obj.browse(cr, uid, uid, context=context)

        author_id = user.partner_id.id

        action = action_obj.browse(cr, uid, action_id, context=context)
        mail_group = action.group_id
        if mail_group:
            u_ids = user_obj.search(cr, uid,
                                    [('group_ids', 'in', mail_group.id)])
            users = user_obj.browse(cr, uid, u_ids)
            partner_ids = map(lambda x: x.partner_id.id, users)
        else:
            partner_ids = []

        if not (uid == action.user_id.id) and not (author_id in partner_ids) \
                and not (uid == action.create_uid.id):
            raise osv.except_osv(
                _("Error in action"),
                _("You cannot mark this action as done because you " +
                  "are not the assgined user of this action."))

        done_date = time.strftime(DSDTF)
        write_vals = {
            'state': 'done',
            'done_date': done_date,
        }
        action_obj.write(cr, uid, action_id, write_vals, context=context)
        return True

    # Funciones del workflow
    def action_confirmed(self, cr, uid, ids, context=None):
        obj_data = self.pool.get('ir.model.data')

        # Creamos los requisitos
        for order in self.browse(cr, uid, ids, context=context):
            # Hacemos chequeos previos
            if order.type.check_mandatory:
                if not order.system_id:
                    raise osv.except_osv(
                        _("Service Order Error!"),
                        _("You cannot confirm this Service Order " +
                          "before you fill System associated."))

            if not order.initial_survey or order.initial_survey.strip() == "":
                raise osv.except_osv(
                    _("Service Order Error!"),
                    _("You have to fill Initial Survey."))

            if not order.sector_id:
                raise osv.except_osv(
                    _("Service Order Error!"),
                    _("You cannot confirm this Service Order " +
                      "before you fill Sector."))

            # Si es un partner con riesgo de cobro,
            # generamos una espera de cliente
            if order.partner_id.kanban_state == 'risk':
                reason_risk_id = obj_data.get_object_reference(
                    cr, uid, 'service', 'reason_risk')[1],
                vals = {
                    'name': _("This customer has been marked as risky"),
                    'reason': reason_risk_id,
                    'order_id': order.id,
                }
                self.pool.get('service.wait').create(
                    cr, uid, vals, context=context)

            order_vals = {
                'state': 'confirmed',
            }

            if order.name == '/':
                order_vals['name'] = self.pool.get('ir.sequence').get(
                    cr, uid, 'service.order') or '/'

            self.write(cr, uid, [order.id], order_vals, context=context)

        return True

    def action_in_progress(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'in_progress'}, context=context)
        return True

    def action_resolved(self, cr, uid, ids, context=None):

        for order in self.browse(cr, uid, ids, context=context):
            if not order.system_id:
                raise osv.except_osv(
                    _("Service Order Error!"),
                    _("You cannot mark as resolved this Service Order " +
                      "before you fill System associated."))

            # Si aplica la carga de componentes y no los tiene cargados
            if not order.component_no_apply and not len(order.component_ids):
                raise osv.except_osv(
                    _("Service Order Error!"),
                    _("You cannot mark as resolved this Service Order " +
                      "before you fill Components associated."))

        for order in self.browse(cr, uid, ids, context=context):
            write_vals = {
                'state': 'resolved',
            }
            if not order.date_resolved:
                write_vals['date_resolved'] = time.strftime(
                    DSDTF)
            self.write(cr, uid, [order.id], write_vals, context=context)
        return True

    def action_quotation(self, cr, uid, ids, context=None):
        return True

    def action_ready(self, cr, uid, ids, context=None):

        self.write(cr, uid, ids, {'state': 'ready'}, context=context)
        self._ready_notify(cr, uid, ids, context=context)
        return True

    def action_customer_confirmed(self, cr, uid, ids, context=None):
        """Funcion llamada desde el workflow al pasar a la activity
           act_customer_confirmed.
        """
        self.write(cr, uid, ids, {'state': 'customer_confirmed'},
                   context=context)
        return True

    def send_materials(self, cr, uid, ids, context=None):
        material_obj = self.pool.get("service.order.material.line")
        move_obj = self.pool.get("stock.move")
        pick_obj = self.pool.get('stock.picking')

        if not context:
            context = {}

        order_data = self.browse(cr, uid, ids[0])

        # Creamos el picking interno como para la valija correspondiente
        pick_vals = {
            'origin': order_data.name,
            'type': 'out',
            'auto_picking': False,
        }

        moves = []
        materials = []
        for material in order_data.material_line_ids:
            if material.state == 'draft':
                if not material.location_id:
                    raise osv.except_osv(
                        _("Stock Error!"),
                        _("You have to define a location to " +
                          "get the materials from"))

                # Creamos el stock move para todos los materiales
                move_id = self._create_stock_move(
                    cr, uid, order_data, material,
                    material.location_id.id, context)
                moves.append((4, move_id))
                materials.append(material.id)
                # Lo ponemos en assigned porque ya hay un warning
                # para el usuario que le avisa
                # que no hay stock, asi que tomamos como que sabe que
                # no deberia mandar los materiales si no hay stock
                move_obj.force_assign(cr, uid, [move_id], context=context)
                write_vals = {
                    'state': 'sent',
                    'move_id': move_id,
                }
                material_obj.write(cr, uid, material.id, write_vals,
                                   context=context)

        # Asociamos los stock_moves al picking y
        # guardamos el picking en las material lines
        pick_vals['move_lines'] = moves
        pick_id = pick_obj.create(cr, uid, pick_vals, context=context)
        pick_obj.draft_force_assign(cr, uid, [pick_id])
        material_obj.write(cr, uid, materials,
                           {'picking_id': pick_id}, context=context)

        return True

    def action_done(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'state': 'done'}, context=context)
        return True

    def action_cancel(self, cr, uid, ids, context=None):
        if type(ids) == int:
          ids = [ids]
        for order in self.browse(cr, uid, ids, context):
          if order.assigned_to_ids:
            raise osv.except_osv(
                          _("Service Order Error!"),
                          _("You can not cancel a service order with assigned employees"))
          for wait in order.wait_ids:
            if wait.state == "waiting":
              raise osv.except_osv(
                          _("Service Order Error!"),
                          _("You can not cancel a service order with waitings"))
          
        self.write(cr, uid, ids, {'state': 'cancel'}, context=context)
        
        return True


    def test_can_resolved(self, cr, uid, ids, context=None):
        wait_obj = self.pool.get('service.wait')
        order_data = self.browse(cr, uid, ids[0])

        # Chequeamos que no hayan esperas que no esten finalizadas aun
        if wait_obj.search(cr, uid, [('order_id', '=', order_data.id),
                                     ('state', '=', 'waiting')]):
            raise osv.except_osv(_('Cannot mark service order as "Ready"'),
                                 _('All waits must be in "finished" state.'))

        # TODO: Todos los materiales deberian estar enviados al menos
        return True

    def test_can_progress(self, cr, uid, ids, context=None):
        return True

    def test_can_ready(self, cr, uid, ids, context=None):
        order_data = self.browse(cr, uid, ids[0])

        # Chequeamos que haya una fecha de asignacion
        if not order_data.assign_date:
            raise osv.except_osv(
                _('Cannot mark service order as "Ready"'),
                _('Field "Assign date" is required.'))

        # Chequeamos que haya gente asignada para hacer el trabajo
        if not order_data.assigned_to_ids:
            raise osv.except_osv(
                _('Cannot mark service order as "Ready"'),
                _('No employees assigned.'))

        # TODO: Falta lo de los materiales
        return True

    def test_quotation(self, cr, uid, ids, context=None):
        # Chequemos que este quoted en True.
        # Si esta en False es porque el usuario
        # todavia no realizo las cotizaciones

        for order in self.browse(cr, uid, ids, context=context):
            # Si el usuario no marco que hizo las cotizaciones,
            # devolvemos False para que no pase al siguiente estado
            if not order.quoted:
                return False

        return True

    def action_service_done(self, cr, uid, ids, context=None):
        # Esta funcion se llama con el boton Finalizar
        wf_service = netsvc.LocalService("workflow")

        res = True
        for order in self.browse(cr, uid, ids, context):
            for sale in order.sale_order_ids:
                if not sale.invoiced:
                    # TODO: Chequear cancelado de sale_order
                    raise osv.except_osv(
                        _("Cannot finish"),
                        _("All quotation should be paid or cancelled"))

            for wait in order.wait_ids:
                if not wait.state == 'finished':
                    raise osv.except_osv(
                        _("Cannot finish"),
                        _("There are active waits"))

        for order in self.browse(cr, uid, ids, context):
            wf_service.trg_validate(uid, 'service.order', order.id,
                                    'service_done', cr)

        return res

    def sale_order_id_service_get(self, cr, uid, ids, context=None):

        sale_ids = []
        for order in self.browse(cr, uid, ids, context=context):
            sale_ids += [s.id for s in order.sale_order_ids]

        return sale_ids

    def mark_quoted(self, cr, uid, ids, context=None):
        self.write(cr, uid, ids, {'quoted': True}, context=context)
        return True

    # Cuando ya no se requieren mas modifiaciones se pasa a estado ready.
    def mark_ready(self, cr, uid, ids, context=None):
        order_data = self.browse(cr, uid, ids[0])

        if not order_data.assign_date:
            raise osv.except_osv(
                _('Cannot mark service order as "Ready"'),
                _('Field "Assign date" is required.'))

        if not order_data.assigned_to_ids:
            raise osv.except_osv(
                _('Cannot mark service order as "Ready"'),
                _('No employees assigned.'))

        # TODO: Revisar esta condicion, porque si el cliente es hard y
        # no tiene ninguna sale.order, entonces no deja pasar a Ready.
        if order_data.partner_id.type == 'hard' and \
                not order_data.sale_order_ids or \
                not self._check_confirmed_sales(order_data.sale_order_ids):
            raise osv.except_osv(
                _('Cannot mark service order as "Ready"'),
                _('At least one invoice is required. ' +
                  'All invoices must be in "confirmed" state.'))
        self.write(cr, uid, ids, {'state': 'ready'}, context)

    def _check_confirmed_sales(self, sale_ids):
        for sale in sale_ids:
            if sale.state != 'confirmed':
                return False
        return True

    def _get_subtype_ids(self, cr, uid, subtypes):
        subtype_ids = []
        for subtype in subtypes:
            if '.' not in subtype:
                subtype = 'mail.%s' % subtype
            ref = self.pool.get('ir.model.data').get_object_reference(
                cr, uid, *subtype.split('.'))
            subtype_id = ref and ref[1] or False
            if subtype_id:
                subtype_ids.append(subtype_id)

        return subtype_ids

    def _ready_notify(self, cr, uid, ids, context=None):

        for order in self.browse(cr, uid, ids, context=context):
            # Buscamos los usuarios asignados
            partner_ids = [assigned.user_id.partner_id.id for assigned
                           in order.assigned_to_ids]

            body = _("Service order planned. Expected Date: %s") % \
                    (order.expected_date or 'N/A')
            for assign in order.assigned_to_ids:
                body += '<br/>' + \
                        _("\n -- Assigned Date: %s\n  Assigned User: %s") \
                        % (assign.date, assign.user_id.name)
            author_id = self.pool.get('res.users').\
                browse(cr, uid, uid, context=context).partner_id.id
            self.message_post(cr, uid, ids, body=body, subject="Planned",
                              type="notification", subtype="mt_comment",
                              context=context, author_id=author_id,
                              partner_ids=partner_ids)
        return True

    # La orden se marca como resuelta.
    def mark_resolved(self, cr, uid, ids, context=None):
        write_vals = {
            'state': 'resolved',
            'date_resolved': time.strftime(DSDTF),
        }
        self.write(cr, uid, ids, write_vals, context)

    def mark_finished(self, cr, uid, ids, context=None):
        if not self.read(cr, uid, ids[0], ['component_ids'])['component_ids']:
            raise osv.except_osv(
                _('Cannot finish service order'),
                _('Field "component_ids" is required'))

        self.write(cr, uid, ids, {'state': 'finished'}, context)

    # En los estados 'ready' y 'resolved' se puede generar una sale order
    # cuyas lineas seran las material lines.
    def generate_sale_order(self, cr, uid, ids, context=None):

        sale_order_obj = self.pool.get('sale.order')
        sale_line_obj = self.pool.get('sale.order.line')
        service_order_data = self.browse(cr, uid, ids[0], context=context)

        lines_to_quote = []
        lines = [
            {
                'product_id': line.product_id.id,
                'name': line.product_id.name,
                'product_uom': line.product_uom.id,
                'product_uom_qty': line.quantity,
                'id': line.id,
            } for line in service_order_data.material_line_used_ids if
            line.state == 'used' and not line.quoted and
            not line.customer_material]

        if not lines:
            return True

        quoted_line_ids = [l['id'] for l in lines]

        # Agrupamos por producto
        for key, group in groupby(lines, lambda x: x['product_id']):
            qty = 0.0
            name = False
            uom_id = False
            for l in group:
                qty += l['product_uom_qty']
                name = l['name']
                uom_id = l['product_uom']

            lines_to_quote.append({
                'name': name,
                'product_id': key,
                'product_uom': uom_id,
                'product_uom_qty': qty
            })

        sale_order_data = {
            'origin': service_order_data.name,
            'partner_id': service_order_data.partner_id.id,
        }

        # Llamamos al onchange_partner_id para cargar datos
        # adicionales que son requeridos por la sale_order.
        onchange_partner_id_vals = sale_order_obj.onchange_partner_id(
            cr, uid, False, sale_order_data['partner_id']
        )
        sale_order_data.update(onchange_partner_id_vals['value'])

        # Llamamos al onchange_user_id para cargar el section_id
        onchange_user_id_vals = sale_order_obj.onchange_user_id(
            cr, uid, [], uid, context=context
        )
        sale_order_data.update(onchange_user_id_vals['value'])

        def convert_to_x2Many_format(self, data_dict):
            with_value = False
            if 'value' in data_dict:
                return_dict = data_dict
                data_dict = data_dict['value']
                with_value = True
            for key, value in data_dict.items():
                if key not in self._columns.keys():
                    raise osv.except_osv(
                        _("Error!"),
                        _("Field %s not in model %s" % (key, self)))
                if self._columns[key].__class__ in (fields.many2many,
                                                    fields.one2many):
                    data_dict[key] = [(6, False, data_dict[key])]
            if with_value:
                return_dict.update(value=data_dict)
            else:
                return_dict = data_dict

            return return_dict

        # Lineas de la cotizacion
        context['ptype'] = 'place'
        for sale_line in lines_to_quote:
            line_data = sale_line_obj.product_id_change(
                cr, uid, False, sale_order_data['pricelist_id'],
                product=sale_line['product_id'],
                qty=sale_line['product_uom_qty'], uom=sale_line['product_uom'],
                qty_uos=0, uos=False, name=sale_line['name'],
                partner_id=sale_order_data['partner_id'],
                lang=False, update_tax=True, date_order=False, packaging=False,
                fiscal_position=sale_order_data['fiscal_position'], flag=False,
                context=context)

            # Como son productos ya retirados de stock e
            # instalados en el cliente el plazo de entrega es 0 y
            # el metodo de abastecimiento es de stock
            line_data['value'].update({
                'delay': 0,
                'type': 'make_to_stock',
            })
            line_data = convert_to_x2Many_format(sale_line_obj, line_data)
            sale_line.update(line_data['value'])

        # Creamos la sale.order y sus lineas
        sale_order_data['order_line'] = [(0, 0, vals) for vals
                                         in lines_to_quote]
        sale_order_id = sale_order_obj.create(
            cr, uid, sale_order_data, context=context)

        # Creamos la referencia a esta sale_order.
        self.write(cr, uid, ids, {'sale_order_ids': [(4, sale_order_id)]},
                   context=context)

        self.pool.get('service.order.material.line.used').write(
            cr, uid, quoted_line_ids, {'quoted': True}, context=context)

        # Creamos una espera por cotizacion
        self.create_wait_for_quotation(cr, uid, service_order_data.id,
                                       context=context)

        return True

    def generate_empty_sale_order(self, cr, uid, ids, context=None):
        sale_order_obj = self.pool.get('sale.order')
        service_order_data = self.browse(cr, uid, ids[0], context=context)

        sale_order_data = {
            'origin': service_order_data.name,
            'partner_id': service_order_data.partner_id.id
        }

        # Llamamos al onchange_partner_id para cargar datos
        # adicionales que son requeridos por la sale_order.
        onchange_partner_id_vals = sale_order_obj.onchange_partner_id(
            cr, uid, False, sale_order_data['partner_id']
        )
        sale_order_data.update(onchange_partner_id_vals['value'])

        # Llamamos al onchange_user_id para cargar el section_id
        onchange_user_id_vals = sale_order_obj.onchange_user_id(
            cr, uid, [], uid, context=context
        )
        sale_order_data.update(onchange_user_id_vals['value'])

        sale_order_id = sale_order_obj.create(
            cr, uid, sale_order_data, context=context)

        # Creamos la referencia a esta sale_order.
        self.write(cr, uid, ids, {'sale_order_ids': [(4, sale_order_id)]},
                   context=context)

        # Creamos una espera por cotizacion
        self.create_wait_for_quotation(cr, uid, service_order_data.id,
                                       context=context)

        return sale_order_id

    def create_wait_for_quotation(self, cr, uid, order_id, context=None):
        obj_data = self.pool.get('ir.model.data')

        reason_quotation_id = obj_data.get_object_reference(
            cr, uid, 'service', 'reason_quotation')[1]

        vals = {
            'name': _("A new quotation was created"),
            'reason': reason_quotation_id,
            'order_id': order_id,
        }

        return self.pool.get('service.wait').create(
            cr, uid, vals, context=context)

    def generate_pickings_out(self, cr, uid, ids, context=None):
        ir_sequence = self.pool.get('ir.sequence')
        picking_obj = self.pool.get('stock.picking')
        stock_move_obj = self.pool.get('stock.move')
        order = self.browse(cr, uid, ids[0], context)

        # Obtenemos el proximo numero en la sequence
        pick_name = ir_sequence.get(cr, uid, 'stock.picking.out')

        picking_vals = {
            'name': pick_name,
            'origin': order.name,
            'type': 'in',
            'move_type': 'one',
            'state': 'auto',
            'auto_picking': True,
            'company_id': order.company_id.id,
        }

        pick_id = picking_obj.create(cr, uid, picking_vals, context=context)

        # Creamos los stock_move para este picking que acabamos de crear
        line_ids = []
        for line in order.material_line_ids:
            # Ubicacion de Origen
            source_location_id = line.location_id.id
            destination_location_id = line.location_dest_id.id
            move_name = _('SERVICE: %s') % line.order_id.name

            line_data = {
                'name': move_name,
                'date': time.strftime('%Y-%m-%d %H:%M:%S'),
                'product_id': line.product_id.id,
                'product_qty': line.quantity,
                'product_uom': line.product_uom.id,
                'product_uos_qty': False,
                'product_uos': False,
                'location_id': source_location_id,
                'location_dest_id': destination_location_id,
                'picking_id': pick_id,
                'prodlot_id': line.component_ids.prodlot_id.id,
                'state': 'waiting',
                'company_id': line.company_id.id,
            }

            move_id = stock_move_obj.create(cr, uid, line_data,
                                            context=context)
            line_ids.append(move_id)

        # En realidad, deberiamos llamar a action_assign de stock.move
        # pero el problema es que nosotros estamos trayendo mercaderia
        # de una ubicacion de cliente y eso no
        # lo toma en cuenta el action_assign,
        # porque de hecho no es una situacion comun. Lo que hacemos es tomar
        # la porcion de codigo de la funcion action_assign que nos sirve y la
        # utilizamos aqui.
        stock_move_obj.write(cr, uid, line_ids, {'state': 'assigned'})

        wf_service = netsvc.LocalService("workflow")
        wf_service.trg_write(uid, 'stock.picking', pick_id, cr)


class NotAssignableTimes(osv.osv):
    _name = 'not.assignable.times'
    _description = "Not Assignable Times"

    def add_all_users(self, cr, uid, ids, context=None):
        user_model = self.pool['res.users']
        users = user_model.search(
            cr, uid, [('category_id', 'ilike', 'Servicio Técnico')],
            context=context)
        form = self.browse(cr, uid, ids[0], context)
        for not_assign_line in form.not_assign_line_ids:
            if not_assign_line.user_id.id in users:
                users.remove(not_assign_line.user_id.id)
        child_obj = self.pool['not.assignable.times.lines']
        for user in users:
            child_obj.create(
                cr, uid, {
                    'user_id': user,
                    'not_assign_time_id': ids[0],
                }, context=context)
        return True

    def remove_all_users(self, cr, uid, ids, context=None):
        child_obj = self.pool['not.assignable.times.lines']
        form = self.browse(cr, uid, ids[0], context)
        unlink_ids = map(lambda x: x.id, form.not_assign_line_ids)
        child_obj.unlink(cr, uid, unlink_ids, context=context)
        return True

    def write(self, cr, uid, ids, vals, context=None):
        if context is None:
            context = {}
        c_vals = {}
        child_obj = self.pool['not.assignable.times.lines']
        form = self.browse(cr, uid, ids[0], context=context)
        child_fields = ['date_start', 'date_end', 'name', 'message']
        for not_assign_line in form.not_assign_line_ids:
            for field in vals.keys():
                if field in child_fields:
                    c_vals[field] = vals[field]
            child_obj.write(cr, uid, [not_assign_line.id], c_vals,
                            context=context)
        return super(NotAssignableTimes, self).write(
            cr, uid, ids, vals, context=context)

    def unlink(self, cr, uid, ids, context=None):
        nats = self.browse(cr, uid, ids, context=context)
        line_ids = []
        for n in nats:
            line_ids += map(lambda x: x.id, n.not_assign_line_ids)

        res = super(NotAssignableTimes, self).unlink(cr, uid, ids, context)

        # Unlink lines manually because of resource.mixin
        child_obj = self.pool['not.assignable.times.lines']
        child_obj.unlink(cr, uid, line_ids, context=context)

        return res

    _columns = {
        'name': fields.char('Name', size=64, required=True),
        'message': fields.char('Notes', size=256),
        'not_assign_line_ids': fields.one2many(
            'not.assignable.times.lines', 'not_assign_time_id',
            string='Not Assign Time Lines', required=True),
        'date_start': fields.datetime('Date Start', required=True),
        'date_end': fields.datetime('Date End', required=True),
        'is_holiday': fields.boolean(string="Is Holiday?"),
    }

    def _check_dates(self, cr, uid, ids, context=None):
        for assignable_time in self.browse(cr, uid, ids, context):
            if assignable_time.date_start > assignable_time.date_end:
                return False
        return True

    _constraints = [
        (_check_dates, 'Error! Date Start must be less '
         'than or equal to the Date End.', ['date_start', 'date_end'])
    ]


class NotAssignableTimesLines(mixin.resource):
    _name = 'not.assignable.times.lines'
    _description = "Not Assignable Times Lines"

    _booking_chart_ref = 'service.assignee_booking_chart'

    _booking_resource_map = {
        'name':         'name',
        'message':      'message',
        'date_start':   'date_start',
        'date_end':     'date_end',
        'resource_ref': 'user_id',
        'target_ref':   'not_assign_time_id',
    }

    def create(self, cr, uid, vals, context=None):
        if context is None:
            context = {}
        parent_obj = self.pool['not.assignable.times']
        parent_id = vals.get('not_assign_time_id', False)
        if parent_id:
            parent = parent_obj.browse(cr, uid, parent_id)
            vals['date_start'] = parent.date_start
            vals['date_end'] = parent.date_end
            vals['message'] = parent.message
            vals['name'] = parent.name
        return super(NotAssignableTimesLines, self).create(
            cr, uid, vals, context)

    _columns = {
        'not_assign_time_id': fields.many2one('not.assignable.times',
                                              'Parent'),
        'name': fields.char(string='Name'),
        'message': fields.char(string='Notes'),
        'user_id': fields.many2one('res.users', 'Employee', required=True),
        'date_start': fields.datetime(string='Date Start'),
        'date_end': fields.datetime(string='Date End'),
    }


# Service Order Assignee
class service_order_assignee(mixin.resource):
    _name = "service.order.assignee"
    _description = "Service Order Assignees"

    # ref to the booking_chart xml_id (if you have created the booking chart
    # manually, you have to override the mixin.resource.get_chart_id method)
    _booking_chart_ref = 'service.assignee_booking_chart'

    _booking_resource_map = {
        # simple mapping, booking.resource field = task field
        'name':        'name',
        'type':        'type',
        'system_name': 'system_name',
        'requirements': 'requirements',
        'message':     'message',
        'date_start':  'date',
        'date_end':    'date_end',
        # object mapping,
        # booking.resource field = "task.field._name,task.field.id"
        'resource_ref': 'user_id',
        'target_ref': 'order_id',
        # custom mapping, set booking.resource.css_class field when
        # priority is updated with the value of task.booking_css_class
        'css_class':   'order_state'
    }

    def _date_end_calc(self, cr, uid, ids, field_name, arg, context=None):
        """ Calcs date_end based on date+hours
        @return: Dictionary of values
        """
        res = {}
        if not ids:
            return res

        for id in ids:
            res[id] = False
            assignee = self.browse(cr, uid, id)
            date_end = datetime.strptime(
                assignee.date, DSDTF) \
                + timedelta(hours=assignee.hours)
            res[id] = date_end.strftime(DSDTF)
        return res

    _columns = {
        'name': fields.char('Name', size=64),
        'type': fields.char('Type', size=256),
        'system_name': fields.char('System Name', size=256),
        'requirements': fields.char('Requirements', size=256),
        'message': fields.char('Notes', size=256),
        'user_id': fields.many2one('res.users', 'Employee', required=True),
        'order_id': fields.many2one('service.order', 'Service Order'),
        'date': fields.datetime('Date', required=True),
        'date_end': fields.datetime('Date end'),
        'order_state': fields.char(size=32, string="Order State"),
        'hours': fields.float('Hours', required=True),
    }

    _defaults = {
        'date': lambda s, cr, uid, c: datetime.strftime(
            datetime.combine(datetime.strptime(
                fields.date.context_today(s, cr, uid, context=c),
                DSDF),
                datetime.min.time()).replace(
                    hour=11, minute=0, second=0) + timedelta(days=1),
            DSDTF),
        'hours': 8.0,
    }

    def _check_not_assignable_time(self, cr, uid, ids, context=None):
        for assignee in self.browse(cr, uid, ids, context=context):
            overlap_query = """
            SELECT natl.id, natl.date_start AS start, natl.date_end AS end
            FROM not_assignable_times_lines natl
            JOIN res_users u ON u.id=natl.user_id
            WHERE (natl.date_start, natl.date_end) OVERLAPS
            (%(start_date)s, INTERVAL '1 hours' * %(hours)s)
            AND u.id=%(user)s
            """
            params = {
                'start_date': assignee.date,
                'hours': assignee.hours,
                'user': assignee.user_id.id,
            }
            cr.execute(overlap_query, params)
            if cr.rowcount:
                return False
        return True

    _constraints = [
        (_check_not_assignable_time,
         '\nError!\nAn Engineer can not be assigned in the selected date.',
         ['user_id', 'date', 'date_end']),
    ]

    def onchange_assignee_line(self, cr, uid, ids, user_id,
                               date, hours, context=None):
        user_obj = self.pool.get('res.users')
        res = {'value': {}}

        # Chequeamos solapamiento
        if not hours or not user_id or not date:
            return res

        user = user_obj.browse(cr, uid, user_id)

        overlap_query = """
        SELECT natl.id, natl.date_start AS start, natl.date_end AS end
        FROM not_assignable_times_lines natl
        JOIN res_users u ON u.id=natl.user_id
        WHERE (natl.date_start, natl.date_end) OVERLAPS
        (%(start_date)s, INTERVAL '1 hours' * %(hours)s)
        AND u.id=%(user)s
        """
        params = {
            'start_date': date,
            'hours': hours,
            'user': user_id,
        }
        cr.execute(overlap_query, params)
        if cr.rowcount:
            raise osv.except_osv(
                _("Error!"),
                _("The Engineer %s can not be assigned the date %s" %
                  (user.name, date)))

        # Buscamos algun tipo de solapamiento de fechas del mismo ingeniero
        overlap_query = """
        SELECT al.id, al.date AS start, al.date + (INTERVAL '1 hour'*al.hours)
        AS end
        FROM service_order_assignee al
        JOIN res_users u ON u.id=al.user_id
        WHERE (al.date, INTERVAL '1 hour'*al.hours)
        OVERLAPS (%s, INTERVAL '1 hours'*%s)
        AND u.id=%s
        """

        params = (date, hours, user_id,)
        cr.execute(overlap_query, params)

        warning_msgs = []
        for overlapped in cr.fetchall():
            date_assigned = datetime.strptime(
                overlapped[1],
                DSDTF).strftime('%d/%m/%Y')
            msg = _("Overlapped assignation for engineer %s. " +
                    "Check Calendar on date %s!") % (user.name, date_assigned)
            warning_msgs.append(msg)

        if warning_msgs:
            warning = {
                'title': _('Warning!'),
                'message': '\n'.join(warning_msgs),
            }

            res['warning'] = warning

        return res

    def write(self, cr, uid, ids, vals, context=None):

        for assignee_reads in self.read(
                cr, uid, ids, ['hours', 'date'], context=context):
            hours = 'hours' in vals and vals['hours'] or \
                assignee_reads['hours']
            date_start = 'date' in vals and vals['date'] or \
                assignee_reads['date']

            if 'hours' in vals or 'date' in vals:
                vals['date_end'] = datetime.strptime(
                    date_start,
                    DSDTF) + timedelta(hours=hours)

        return super(service_order_assignee, self).write(
            cr, uid, ids, vals, context)

    def create(self, cr, uid, vals, context=None):

        order_obj = self.pool.get('service.order')

        if vals['order_id']:
            order_id = vals['order_id']
            reads = order_obj.read(cr, uid, order_id,
                                   ['partner_id', 'state', 'system_id',
                                    'requirement_ids', 'type'])

            customer_name = reads['partner_id'][1]
            state = reads['state']

            # Estos campos los tenemos que calcular aca porque sino el
            # booking.mixin no crea los bookings.
            # TODO: Llamar a la misma funcion del campo calculado
            reqs = ''
            if reads['requirement_ids']:
                requirement_model = self.pool['service.requirement']
                requirements = requirement_model.browse(
                    cr, uid, reads['requirement_ids'], context=context)
                first = True
                for requirement in requirements:
                    if first:
                        reqs += (requirement.name)
                        first = False
                    else:
                        reqs += (' - ' + requirement.name)
            vals['name'] = customer_name
            vals['type'] = reads['type'][1] if reads['type'] else ''
            vals['system_name'] = reads['system_id'][1] if reads['system_id'] \
                else ''
            vals['requirements'] = reqs
            vals['order_state'] = service_order._colors_mapping[state]

        vals['date_end'] = datetime.strptime(
            vals['date'], DSDTF) \
            + timedelta(hours=vals['hours'])

        return super(service_order_assignee, self).create(
            cr, uid, vals, context=context)

    def name_get(self, cr, uid, ids, context=None):
        if not context:
          context = {}

        res = []
        for soa in self.browse(cr, uid, ids, context=context):
            tz = context.get('tz', False)
            local = tz and pytz.timezone(tz) or pytz.utc
            date_utc = datetime.strptime(soa.date,
                                         DSDTF)
            date_local = pytz.utc.localize(date_utc).astimezone(local)
            str_date_local = datetime.strftime(date_local,
                                               DSDTF)
            name = soa.user_id.name + ' - ' + str_date_local
            res.append((soa.id, name,))
        return res


# Service Order Lines
class service_order_material_line(osv.osv):
    _name = "service.order.material.line"
    _description = "Materials"
    _columns = {
        'order_id': fields.many2one('service.order', 'Service Order'),
        'product_id': fields.many2one(
            'product.product', 'Product',
            domain=[('product_tmpl_id.type', '!=', 'service')],
            readonly=True, states={'draft': [('readonly', False)]}),
        'product_uom': fields.many2one(
            'product.uom', 'Unit of Measure ', required=True,
            readonly=True, states={'draft': [('readonly', False)]}),
        'quantity': fields.integer('Quantity', readonly=True,
                                   states={'draft': [('readonly', False)]}),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse'),
        'location_id': fields.many2one(
            'stock.location', 'Location', select=True, readonly=True,
            states={'draft': [('readonly', False)]},
            domain=[('usage', '=', 'internal')]),
        'state': fields.selection([
            ('draft', 'Draft'),
            ('sent', 'Sent'),
            ('used', 'Used'),
            ('scrap', 'Scrap'),
            ('returned', 'Returned'),
            ('cancel', 'Cancelled')], 'State', readonly=True),
        'move_id': fields.many2one('stock.move'),
        'picking_id': fields.many2one('stock.picking', 'Picking'),
        'company_id': fields.many2one('res.company', 'Company'),
        'quoted': fields.boolean('Quoted'),  # Sirve para no volver a cotizar
    }

    _defaults = {
        'state': lambda *a: 'draft',
        'company_id': lambda self, cr, uid, context: self.pool.get(
            'res.company')._company_default_get(
                cr, uid, 'res.partner', context=context),
        'quoted': False,
    }

    def onchange_product_id(self, cr, uid, ids, product_id,
                            quantity, product_uom, context=None):
        product_obj = self.pool.get('product.product')
        product_uom_obj = self.pool.get('product.uom')

        result = {}

        if not product_id:
            return {}

        product = product_obj.browse(cr, uid, product_id, context=context)
        if not product_uom:
            result['product_uom'] = product.uom_id.id
            product_uom = product.uom_id.id

        product = product_obj.browse(cr, uid, product_id, context=context)

        # warning_msgs = ''
        # warning = {}

        uom2 = False
        if product_uom:
            uom2 = product_uom_obj.browse(cr, uid, product_uom,
                                          context=context)
            if product.uom_id.category_id.id != uom2.category_id.id:
                product_uom = product.uom_id.id
                result['product_uom'] = product.uom_id.id
        if not uom2:
            uom2 = product.uom_id

        # compare_qty = float_compare(product.virtual_available, quantity,
        #                             precision_rounding=uom2.rounding)

        # if product.type == 'product' and int(compare_qty) == -1 and \
        #         product.procure_method == 'make_to_stock':
        #     warn_msg = _('You plan to sell %.2f %s but you only have ' +
        #                  '%.2f %s available !\nThe real stock is %.2f %s. ' +
        #                  '(without reservations)') % \
        #         (quantity, uom2.name,
        #          max(0, product.virtual_available), uom2.name,
        #          max(0, product.qty_available), uom2.name)
        #
        #     warning_msgs += _("Not enough stock ! : ") + warn_msg + "\n\n"
        #
        #     warning = {
        #        'title': _('Stock Error!'),
        #        'message': warning_msgs
        #     }

        return {
            'value': result,
            # 'warning': warning,
        }

    def onchange_warehouse_id(self, cr, uid, ids, warehouse_id, context=None):
        warehouse_obj = self.pool.get('stock.warehouse')

        result = {}

        if not warehouse_id:
            return {}

        warehouse = warehouse_obj.browse(cr, uid, warehouse_id,
                                         context=context)
        result['location_id'] = warehouse.lot_stock_id.id


        return {'value': result}

    def unlink(self, cr, uid, ids, context=None):

        if not context.get('force_unlink', False):
            for line in self.browse(cr, uid, ids, context=context):
                if line.state != 'draft':
                    raise osv.except_osv(
                        _("Resources Error!"),
                        _("Cannot delete material resources that has " +
                          "been sent, used, scrapped or returned"))

        super(service_order_material_line, self).unlink(
            cr, uid, ids, context=context)


class service_order_material_line_used(osv.osv):
    _name = "service.order.material.line.used"
    _inherit = "service.order.material.line"

    _columns = {
        'suitcase_id': fields.many2one('stock.suitcase', 'Suitcase',
                                       readonly=True,
                                       states={
                                           'draft': [('readonly', False)]
                                       }),
        'customer_material': fields.boolean('Provided by customer'),
        'no_stock': fields.boolean('Avoid Stock'),
    }

    _defaults = {
        'state': 'draft',
        'customer_material': False,
    }

    def onchange_product_id(self, cr, uid, ids, product_id, quantity,
                            product_uom, context=None):
        product_obj = self.pool.get('product.product')
        product_uom_obj = self.pool.get('product.uom')

        result = {}

        if not product_id:
            return {}

        product = product_obj.browse(cr, uid, product_id, context=context)
        if not product_uom:
            result['product_uom'] = product.uom_id.id
            product_uom = product.uom_id.id

        product = product_obj.browse(cr, uid, product_id, context=context)

#        warning_msgs = ''
#        warning = {}

        uom2 = False
        if product_uom:
            uom2 = product_uom_obj.browse(cr, uid, product_uom,
                                          context=context)
            if product.uom_id.category_id.id != uom2.category_id.id:
                product_uom = product.uom_id.id
                result['product_uom'] = product.uom_id.id
        if not uom2:
            uom2 = product.uom_id

        return {'value': result}

    def create(self, cr, uid, vals, context=None):
        pick_obj = self.pool.get('stock.picking')
        order_obj = self.pool.get("service.order")

        customer_material = 'customer_material' in vals and \
            vals['customer_material'] or False
        manual_created = 'state' not in vals or \
            vals['state'] != 'used' or False
        no_stock = 'no_stock' in vals and vals['no_stock'] or False
        if customer_material or no_stock:
            vals['state'] = 'used'

        material_id = super(service_order_material_line_used, self).\
            create(cr, uid, vals, context)

        # Si el material no fue proveido por el cliente,
        # hacemos el remito interno
        if not customer_material and not no_stock and manual_created:

            # Si fue creado manualmente
            moves = []
            order = order_obj.browse(cr, uid, vals['order_id'],
                                     context=context)
            # Creamos el picking interno
            pick_vals = {
                'origin': order.name,
                'type': 'out',
                'auto_picking': True,
            }

            # Creamos el movimiento de stock correspondiente
            material = self.browse(cr, uid, material_id, context)
            move_id = order_obj._create_stock_move(
                cr, uid, order, material,
                material.suitcase_id.location_id.id, context)
            moves.append((4, move_id))

            # Asociamos los stock_moves al picking y guardamos
            # el picking en las material lines
            pick_vals['move_lines'] = moves
            pick_id = pick_obj.create(cr, uid, pick_vals, context=context)
            self.write(
                cr, uid, material_id, {
                    'picking_id': pick_id,
                    'state': 'used'}, context=context)
            pick_obj.draft_force_assign(cr, uid, [pick_id])
            pick_obj.force_assign(cr, uid, [pick_id])

        return material_id

    def cancel_picking(self, cr, uid, ids, context=None):
        picking_obj = self.pool['stock.picking']
        for material in self.browse(cr, uid, ids, context):
            picking_obj.action_cancel(cr, uid, [material.picking_id.id],
                                      context=context)
            self.write(cr, uid, [material.id], {'state': 'cancel'})


# Service Order Type
class service_order_type(osv.osv):
    _name = "service.order.type"
    _description = "Service Order Type"
    _columns = {
        'name': fields.char('Name', size=32, required=True),
        'check_mandatory': fields.boolean('System  mandatory filling',
                                          help="Check this if system is mandatory fields to confirm Service Order"),  # noqa
        'id_tiposervicio': fields.integer('Id CM'),
    }
