# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (c) 2017 E-MIPS (http://www.e-mips.com.ar)
#    Copyright (c) 2017 Eynes (http://www.eynes.com.ar)
#    All Rights Reserved. See AUTHORS for details.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from openerp.osv import fields, osv
from openerp.tools.translate import _
from datetime import datetime
from dateutil.relativedelta import relativedelta

class sale_order_create_service(osv.osv_memory):
    _name = "sale.order.create.service"
    _description = "Create Service Order from SO"

    _columns = {
        'expected_date': fields.date('Expected date'),
        'type': fields.many2one('service.order.type', 'Type'),
        'times': fields.integer('Times', required=True,
            help="Quantity of Service Orders to Create"),
        'every': fields.integer('Every', required=True),
        'period': fields.selection(
            [('days', 'Days'), ('months', 'Months')],
            'Periodicity', required=True), 
        #'partner_id': fields.many2one('res.partner', 'Partner',
            #domain=[('parent_id', '=', False), ('is_company', '=', True)], required=True),
    }
 
    _defaults = {
        'times': lambda *x: 1,
        'period': 'months',
    }

    def _prepare_service_order(self, cr, uid, expected_date, service_type, partner_id,
            sector_id, employee_sector_id, sale_order_id=None, context=None):

        sale_order_ids = []
        if sale_order_id:
            sale_order_ids.append((4, sale_order_id, 0))

        return {
            'name': "/",
            'type': service_type,
            'partner_id': partner_id,
            'expected_date': expected_date,
            'initial_survey': "",
            'sector_id': sector_id,
            'employee_sector_id': employee_sector_id,
            'sale_order_ids': sale_order_ids, 
        }

    def create_service_orders(self, cr, uid, ids, context=None):

        sale_obj = self.pool.get('sale.order')
        service_obj = self.pool.get('service.order')

        active_ids = context.get('active_ids')
        if not active_ids:
            raise osv.except_osv(_("Error"), _("At least one register has to be selected."))

        # Obtenemos el wizard
        wiz = self.browse(cr, uid, ids[0], context)

        # Obtenemos la SO
        order = sale_obj.browse(cr, uid, active_ids[0], context)
        partner = order.partner_id
        sector_id = order.sector_id.id
        employee_sector_id = order.technique_employee_id.id
        service_type = wiz.type.id
        expected_date = wiz.expected_date

        # Obtenemos las diferentes fechas
        service_ids = []
        for i in range(0, wiz.times):
            expected_date = datetime.strptime(wiz.expected_date, '%Y-%m-%d')

            if wiz.period == 'days':
                new_expected_date = expected_date+relativedelta(days=i*wiz.every)
            else:
                new_expected_date = expected_date+relativedelta(months=i*wiz.every)

            # Creamos la service order
            service_vals = self._prepare_service_order(cr, uid, new_expected_date, service_type,
                    partner.id, sector_id, employee_sector_id, sale_order_id=order.id)

            service_ids.append(service_obj.create(cr, uid, service_vals, context))

        # TODO: Retornar la action con el tree con los service.order
        return service_ids
