from osv import osv, fields
from tools.translate import _
import netsvc
from datetime import datetime

class service_material_quantity(osv.osv_memory):
    _name = "service.material.quantity"
    _description = "Material quantity"
    

    _columns = {
        'scrap_location_id': fields.many2one('stock.location', 'Scrap Location', domain=[('scrap_location','=',True)], required=True),
        'line_ids': fields.one2many('service.material.quantity.line', 'wiz_id', 'Lines')
    }

    def _material_line_get(self, cr, uid, line):

        return {
            'material_line_id': line.id,
            'product_id': line.product_id.id,
            'qty': line.quantity,
            'uom_id': line.product_uom.id,
            'qty_remaining': line.quantity,
            'qty_used': 0,
            'qty_returned': 0,
            'qty_scrap': 0,
        }

    def default_get(self, cr, uid, fields, context=None):
        if context is None: context = {}
        # no call to super!
        res = {}
        order_id = context.get('active_id', [])
        if not order_id or not context.get('active_model') == 'service.order':
            return res
        if 'line_ids' in fields:
            order = self.pool.get('service.order').browse(cr, uid, order_id, context=context)

            lines = [self._material_line_get(cr, uid, l) for l in order.material_line_ids if l.state in ('sent')]
            res.update(line_ids=lines)
#        if 'date' in fields:
#            res.update(date=time.strftime(DEFAULT_SERVER_DATETIME_FORMAT))


        # Scrap Location
        location_id = False
        location_ids = self.pool.get('stock.location').search(cr, uid, [('scrap_location','=',True)], limit=1)
        if location_ids:
            location_id = location_ids[0]
            res.update({'scrap_location_id': location_id})

        return res

    def done(self, cr, uid, ids, context=None):
        line_obj = self.pool.get("service.material.quantity.line")
        material_line_obj = self.pool.get('service.order.material.line')
        picking_obj = self.pool.get('stock.picking')
        move_obj = self.pool.get('stock.move')

        # Valores del wizard
        wiz = self.browse(cr, uid, ids[0], context=context)

        # Chequeo de cantidad negativas por lineas, hacemos el chequeo recalculando porque 
        # remaining es un campo readonly con un on_change que lo modifica, por lo tanto, no se termina de escribir
        for line in wiz.line_ids:
            res = line_obj.qty_change(cr, uid, ids, line.qty, line.qty_used, line.qty_returned, line.qty_scrap)

            # Chequeo de cantidades negativas
            if res and res['value'] and res['value']['qty_remaining'] < 0.0:
                raise osv.except_osv(_("Materials Error!"), _("You have informed more materials than you have. Remaining cannot be negative. Check quantities!."))

            qty_remaining = res['value']['qty_remaining']

            # Obtenemos la material_line original
            material_line = line.material_line_id

            material_vals = {
                'quantity': qty_remaining,
            }

            original_move = material_line.move_id
            if not original_move:
                continue

            # Logica para la cantidad retornada
            if line.qty_returned:
                self._inform_returned_materials(cr, uid, line, material_line, context=context)

            # Logica para la cantidad rota
            if line.qty_scrap:
                self._inform_scrapped_materials(cr, uid, line, material_line, wiz.scrap_location_id.id, context=context)

            # Logica para la cantidad utilizada
            if line.qty_used:
                self._inform_used_materials(cr, uid, line, material_line, context=context)

            # Si se termino de informar correctamente todo y no tenemos cantidad
            # remanente, damos por finalizado el picking
            if qty_remaining == 0.0:
                context['force_unlink'] = True
                context['call_unlink'] = True
                pick_id = original_move.picking_id.id
                material_line_obj.unlink(cr, uid, [material_line.id], context=context)

                move_obj.action_cancel(cr, uid, [original_move.id], context=context)
                move_obj.unlink(cr, uid, [original_move.id], context=context)

                wf_service = netsvc.LocalService("workflow")
                picking_obj.action_move(cr, uid, [pick_id], context=context)
                wf_service.trg_validate(uid, 'stock.picking', pick_id, 'button_done', cr)
            else: 
                # Escribimos el remaining con nueva cantidad de la linea
                move_obj.write(cr, uid, original_move.id, {'product_qty': qty_remaining}, context=context)
                material_line_obj.write(cr, uid, material_line.id, material_vals, context=context)
 

        return {'type': 'ir.actions.act_window_close'}


    def _inform_used_materials(self, cr, uid, line, material_line, context=None):

        partial_obj = self.pool.get('stock.partial.move')

        if not context:
            context = {}
        context['from_service'] = True
        
        context['active_ids'] = [material_line.move_id.id]
        context['active_model'] = 'stock.move'
        partial_id = partial_obj.create(cr, uid, {}, context=context)

        partial = partial_obj.browse(cr, uid, partial_id) 

        # Buscamos el move_id que tenemos que enviar parcial/totalmente segun
        # el campo de qty_used
        if partial.move_ids[0].move_id == material_line.move_id:

            self.pool.get('stock.partial.move.line').write(cr, uid, partial.move_ids[0].id, {'quantity': line.qty_used}, context=context)
            # Confirmamos el movimiento de stock parcial/total
            partial_obj.do_partial(cr, uid, [partial_id], context=context)
            
            # Creamos la nueva linea de material que refleje este cambio de estado
            self._create_new_material_line_used(cr, uid, material_line, line.qty_used, context=context)

        return True

    def _inform_returned_materials(self, cr, uid, line, material_line, context=None):
        move_obj = self.pool.get('stock.move')

        # Realizamos la linea de material correspondiente
        self._create_new_material_line(cr, uid, material_line, 'returned', line.qty_returned, context=context)

        # Si vuelven materiales, lo que hacemos es crear un stock.move para cancelar
        location_dest_id = material_line.move_id.location_dest_id.id
        move_id = self._create_stock_move(cr, uid, material_line.picking_id.id, material_line.order_id, line, line.qty_returned,
                                          material_line.location_id.id, location_dest_id,
                                          desc="Returned", context=context)

        # Cancelamos el movimiento, como para que quede en stock
        move_obj.action_cancel(cr, uid, [move_id], context=context)

        return True

    def _inform_scrapped_materials(self, cr, uid, line, material_line, scrapped_location_id, context=None):
        move_obj = self.pool.get('stock.move')

        # Si se rompen o pierden los materiales, lo que hacemos es crear un stock.move
        # al reves del original pero que la ubicacion destino sea la seleccionada como scrap
        move_id = self._create_stock_move(cr, uid, material_line.picking_id.id, material_line.order_id, line, line.qty_scrap,
                                          material_line.location_id.id, scrapped_location_id,
                                          desc="Scrapped", context=context)

        # Confirmamos y hacemos el movimiento de stock
        move_obj.force_assign(cr, uid, [move_id], context=context)
        move_obj.action_done(cr, uid, [move_id], context=context)

        # Realizamos la linea de material correspondiente
        self._create_new_material_line(cr, uid, material_line, 'scrap', line.qty_scrap, context=context)
        return True

    def _create_stock_move(self, cr, uid, pick_id, order, line, qty, location_id, location_dest_id, desc, context):
        move_obj = self.pool.get('stock.move')

        data = {
            'name': _('Service Order %s (%s)') % (order.name, desc),
            'product_id': line.product_id.id,
            'product_qty': qty,
            'product_uom': line.uom_id.id,
            'location_id': location_id,
            'location_dest_id': location_dest_id,
            'picking_id': pick_id,
            'state': 'confirmed',
        }

        move_id = move_obj.create(cr, uid, data, context=context)

        return move_id

    def _create_new_material_line(self, cr, uid, material_line, state, quantity, context=None):
        material_line_obj = self.pool.get('service.order.material.line')

        vals = {'state': state, 
                'quantity': quantity, 
                'order_id': material_line.order_id.id, 
                'product_id': material_line.product_id.id, 
                'product_uom': material_line.product_uom.id, 
                'location_id': material_line.location_id.id, 
                'warehouse_id': material_line.warehouse_id.id, 
                'picking_id': material_line.picking_id.id,
                #'location_dest_id': material_line.location_dest_id.id
                #'suitcase_id': material_line.suitcase_id.id
        }
        new_line_id = material_line_obj.create(cr, uid, vals, context=context)

        return new_line_id

    def _create_new_material_line_used(self, cr, uid, material_line, quantity, context=None):
        material_line_used_obj = self.pool.get('service.order.material.line.used')

        vals = {'quantity': quantity, 
                'order_id': material_line.order_id.id, 
                'product_id': material_line.product_id.id, 
                'product_uom': material_line.product_uom.id, 
                'picking_id': material_line.move_id.picking_id.id,
                'warehouse_id': material_line.warehouse_id.id, 
                'state': 'used',
        }
        new_line_id = material_line_used_obj.create(cr, uid, vals, context=context)

        return new_line_id


service_material_quantity()

class service_material_quantity_line(osv.osv_memory):
    _name = "service.material.quantity.line"
    _description = "Material quantity line"
    
    _columns = {
        'material_line_id': fields.many2one('service.order.material.line', 'Material Line'),
        'product_id': fields.many2one('product.product', 'Product'),
        'qty': fields.float('Quantity'),
        'qty_remaining': fields.float('Quantity Remaining'),
        'uom_id': fields.many2one('product.uom', 'Unit of Measure'),
        'wiz_id': fields.many2one('service.material.quantity'),
        'qty_used': fields.float('Quantity Used'),
        'qty_returned': fields.float('Quantity Returned'),
        'qty_scrap': fields.float('Quantity Scrap'),
    }

    def qty_change(self, cr, uid, ids, qty, qty_used, qty_returned, qty_scrap):

        qty_remaining = qty

        qty_remaining -= qty_used
        qty_remaining -= qty_returned
        qty_remaining -= qty_scrap
        return {'value': {'qty_remaining': qty_remaining}}



service_material_quantity_line()
