# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (c) 2014 E-MIPS/Eynes All Rights Reserved.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from openerp.osv import osv, fields
from openerp.tools.translate import _
from openerp import netsvc

class service_order_phone_resolve(osv.osv_memory):
    _name = "service.order.phone.resolve"
    _description = "Service Order resolve by phone"
    
    def _get_default_employee_id(self, cr, uid, context=None):
        hr_employee_obj = self.pool['hr.employee']
        employee_id = hr_employee_obj.search(cr, uid, [('user_id', '=', uid)],
                                             context=context)
        return employee_id[0]

    _columns = {
        'create_quotation': fields.boolean('Create quotation', help='Check the box if you want to create a quotation.'),
        'notes': fields.text('Call Summary', required=True),
        'finished_by_id': fields.many2one('hr.employee',
                                          string='Finished by', required=True),
        'finish_id': fields.many2one('service.finish.type',
                                     string='Finish Type', required=True),
    }
    _defaults = {
        'create_quotation': False,
        'finished_by_id': _get_default_employee_id,
    }

    def resolve(self, cr, uid, ids, context=None):
        order_obj = self.pool.get('service.order')
        mod_obj = self.pool.get('ir.model.data')
        act_obj = self.pool.get('ir.actions.act_window')
        wf_service = netsvc.LocalService('workflow')

        sale_id = None

        # Obtenemos datos del wizard
        data = self.browse(cr, uid, ids[0], context=context)
        order_id = context.get('active_id', False)

        # Si se tildo crear cotizacion, la creamos
        if data.create_quotation:
            sale_id = order_obj.generate_empty_sale_order(cr, uid, [order_id], context=context)

        # Avanzamos en el workflow
        wf_service.trg_validate(uid, 'service.order', order_id, 'resolved_phone', cr)

        # Escribimos el texto de la llamada
        final_diagnostic = order_obj.read(cr, uid, order_id, ['final_diagnostic'], context=context)['final_diagnostic']

        summary = _("Call summary: %s") % data.notes
        if final_diagnostic:
            summary = final_diagnostic + '\n' + summary
        vals = {
            'resolved_by_phone': True,
            'final_diagnostic': summary,
            'finished_by_id': data.finished_by_id and data.finished_by_id.id,
            'finish_id': data.finish_id and data.finish_id.id,
        }

        # Escribmos el check de resuelto por telefono
        order_obj.write(cr, uid, order_id, vals, context=context)

        if not sale_id:
            return True

        # Retornamos la vista de tree de las cotizaciones
        result = mod_obj.get_object_reference(cr, uid, 'sale', 'action_quotations')
        id = result and result[1] or False
        result = act_obj.read(cr, uid, [id], context=context)[0]
        result['domain'] = "[('id','=', " + str(sale_id) + ")]"

        return result

service_order_phone_resolve() 
